//-------+---------+---------+---------+---------+---------+---------+--------=
//
// File: grandom.c									Author: Bob Harris
//
// Generic C application program
//
//----------

//----------
//
// grandom--
//  Choose a set of random subranges from a genome.  Actually, you can choose
//  from any set of ranges;  a genome is just one such set.
//
// typical command line (all on one line)
//  grandom human.ranges 100 -s 5 25 -fasta
//       -d /depot/data1/cache/gp/hum/2003-07-01/
//
//----------
//
// Potential Improvements
// ======================
//
// - Currently the allowed intervals are stored in a simple list, as are the
//   generated intervals.  This is *not* an optimal data structure for checking
//   for overlaps.  A better scheme would be to store the allowed intervals in
//   something like a binary tree, and keep this up-to-date by removing (and
//   splitting) intervals each time we generate an interval.  This would also
//   improve the way excluded intervals are handled.
//
// - Performance would be improved by generating all the intervals first,
//   without checking nucleotides;  then sort by chromosome and read by
//   chromosome;  then check.  The process would be repeated for whatever
//   intervals were still needed.  An excess could be generated before checking,
//   to reduce the chance that additional checking passes would be needed.
//
// - maxAttempts should be a command line argument, and there should be some
//   reporting to the user when the number of attempts is beginning to
//   increase, so the user can get a feel for having asked for nearly too
//   much.
//
//----------

static const char rcsid[]=
"$Id: grandom.c,v 1.18 2005/07/26 17:32:39 rsharris Exp rsharris $";

//----------
//
// other files--
//
//----------

#include <stdlib.h>				// standard C stuff
#define  true  1
#define  false 0
#include <stdio.h>				// standard C i/o stuff
#include <string.h>				// standard C string stuff
#include <ctype.h>				// standard C upper/lower case stuff
#include <time.h>				// standard C time stuff
#include "rand32.h"				// Bob's portable random number stuff
#include "seq.h"				// Scott Schwartz's sequence stuff
#include "ranges.h"				// Bob's range list stuff
#include "intvec.h"				// integer vector stuff
#include "util.h"				// Scott Schwartz's handy utility stuff
#include "bobutil.h"			// Bob's handy utility stuff

char* argv0; // (just to satisfy link with piplib)

//----------
//
// global data--
//
//----------

// options

int verbosity = 0;				// how much info to bombard the user with
								//   0  => only the essentials
								//   10 => everything
								//   20 => private debug features
int subrangeLength;
int numSubranges;
int randomStrands;
int allowOverlap;
int allowBadNucs;
int idTags           = false;   // true => assign a unique id tag to each
								//         .. subrange

unsigned long randomSeed;

int maxAttempts  = 100;			// maximum number of attempts to make to
								// .. generate one subrange, before giving up
int charsPerLine = 50;			// number of seq chars to print on one line

// genome (this is just a list of ranges)

char* genomeName  = NULL;		// name of genome ranges file
char* genomeDir   = NULL;		// name of genome nucleotides directory
char* genomeExt   = NULL;		// extension for genome nucleotides files
char* excludeName = NULL;		// name of ranges-to-exclude file
char* lengthsName = NULL;		// name of lengths file
char* fastaName   = NULL;		// name of file to write fasta to
char* seedOutName = NULL;		// name of file to write seed to

//----------
//
// prototypes for local functions--
//
//----------

int main (int argc, char** argv);

static void generate          (ranges_t* r, intvec_t* lengths);
static int  valid_sequence    (SEQ* seq);
static void print_as_fasta    (FILE* f, SEQ* seq);
static void create_lengths    (int howMany, int length, intvec_t* lengths);
static void ranges_to_lengths (ranges_t* r, intvec_t* lengths);
static void read_lengths      (FILE* f, intvec_t* lengths);
static void shuffle_lengths   (intvec_t* lengths);
static void usage             (void);
static void get_args          (int argc, char** argv);
static void show_options      (FILE* f);

//----------
//
// grandom--
//  Main program
//
//----------

int main
   (int			argc,
	char**		argv)
	{
	ranges_t	genome  = ranges_INIT (ckrealloc, ckfree);
	intvec_t	lengths = intvec_INIT (ckrealloc, ckfree);

	//////////
	// fetch arguments
	//////////

	get_args (argc, argv);
	if (verbosity >= 8) show_options (stderr);

	//////////
	// get the allowable intervals
	//////////

	// open and read in the allowed portion of the genome

	if (genomeName == NULL)
		ranges_read (stdin, &genome);
	else
		{
		FILE* genomeF;

		genomeF = ckopen (genomeName, "rt");
		ranges_read (genomeF, &genome);
		fclose (genomeF);  genomeF = NULL;
		}

	// open, read in, and apply the disallowed portion of the genome

	if (excludeName != NULL)
		{
		FILE*		excludeF;
		ranges_t	exclude = ranges_INIT (ckrealloc, ckfree);

		excludeF = ckopen (excludeName, "rt");
		ranges_read (excludeF, &exclude);
		fclose (excludeF);  excludeF = NULL;

		ranges_sort    (&genome,  NULL);
		ranges_sort    (&exclude, NULL);
		ranges_exclude (&genome, &exclude);

		ranges_fini (&exclude);
		}

	// show 'em

	if (verbosity >= 9)
		{
		fprintf        (stderr, "genome\n");
		fprintf        (stderr, "-------\n");
		ranges_fprintf (stderr, &genome, "%s %lu %lu %s %s\n");
		fprintf        (stderr, "\n");
		}

	else if (verbosity >= 8)
		fprintf (stderr, "(%d genomic pieces)\n", genome.len);

	//////////
	// get the lengths
	//////////

	if (lengthsName == NULL)
		create_lengths (numSubranges, subrangeLength, &lengths);
	else
		{
		FILE*		lengthsF;
		ranges_t	lengthsR = ranges_INIT (ckrealloc, ckfree);

		lengthsF = ckopen (lengthsName, "rt");
		read_lengths (lengthsF, &lengths);

		if (lengths.len == 0)
			{
			rewind (lengthsF);

			ranges_read (lengthsF, &lengthsR);
			fclose (lengthsF);  lengthsF = NULL;

			if (lengthsR.len < 1)
				fatalf ("%s contains no example lengths", lengthsName);

			ranges_to_lengths (&lengthsR, &lengths);

			ranges_fini (&lengthsR);
			}
		}

	//////////
	// generate sequences
	//////////

	generate (&genome, &lengths);

	//////////
	// clean up
	//////////

	ckfree (genomeName);
	ckfree (genomeDir);
	ckfree (genomeExt);
	ckfree (excludeName);
	ckfree (lengthsName);
	ckfree (fastaName);
	ckfree (seedOutName);

	ranges_fini (&genome);
	intvec_fini (&lengths);

	return 0;
	}

//----------
//
// generate--
//  Generate a set of random subranges.
//
//----------
//
// Arguments:
//	ranges_t*	r:			The range list to choose subranges from.
//	intvec_t*	lengths:	Lengths of the desired subintervals.  It is
//							.. expected, but not checked, that all the lengths
//							.. are positive.  Note that as a side effect, this
//							.. list gets shuffled.
//
// Returns:
//  (nothing)
//
//----------

static void generate
   (ranges_t*		r,
	intvec_t*		lengths)
	{
	SEQ*			seq         = NULL;
	char*			seqFileName = NULL;
	FILE*			fastaF      = NULL;
	range_t*		rng;
	range_t			sub;
	ranges_t		subranges = ranges_INIT (ckrealloc, ckfree);
	unsigned long   minLen, maxLen, rngLen, totalLen, subLen;
	unsigned long   places, totalPlaces, start, saveStart;
	int				subNum, idNum, attempts;
	int				ix;
	char*			idTag = NULL;

	//////////
	// open output files
	//////////

	if (fastaName != NULL)
		fastaF = ckopen (fastaName, "wt");

	if (seedOutName != NULL)
		{
		FILE* seedF = ckopen (seedOutName, "wt");
		fprintf (seedF, "%ld\n", randomSeed);
		fclose  (seedF);
		}

	//////////
	// shuffle the lengths
	//////////

	// determine range of lengths

	if (lengths->len < 1)
		return;

	minLen = maxLen = lengths->a[0];

	for (subNum=1; subNum<lengths->len ; subNum++)
		{
		subLen = lengths->a[subNum];
		if      (subLen < minLen) minLen = subLen;
		else if (subLen > maxLen) maxLen = subLen;
		}

	// shuffle the lengths;  we do this to distribute any potential length
	// order bias that might be inherent in the interval selection process
	// (we don't shuffle if the list is homogenous, because we want avoid using
	// up random numbers, so we can preserve backward compatibility with sets
	// generated from a specific seed by earlier versions of the program that
	// lacked length distribution capability)

	if (minLen != maxLen)
		shuffle_lengths (lengths);

	//////////
	// calculate the number of possible starting locations;  note that we do
	// this only for the minimum length interval to be generated;  for longer
	// lengths we will reject some intervals for crossing the end of one of the
	// allowed ranges
	//////////

	totalPlaces = 0;
	totalLen    = 0;

	for (ix=0 ; ix<r->len ; ix++)
		{
		rng = &r->vec[ix];
		rngLen = (rng->end+1) - rng->start;
		totalLen += rngLen;
		if (rngLen < minLen) continue;
		places      =  rngLen+1 - minLen;
		totalPlaces += places;
		}

	if (totalPlaces == 0)
		fatalf ("there are no subranges of length %d", minLen);

	if (verbosity >= 8)
		fprintf (stderr, "(%lu bases, %lu possible length %lu subranges)\n",
		         totalLen, totalPlaces, minLen);

	//////////
	// generate the subranges
	//////////

	idNum = 1;

	for (subNum=0 ; subNum<lengths->len ; subNum++)
		{
		subLen = lengths->a[subNum];

		if (verbosity >= 20)
			fprintf (stderr, "generating length #%d (%lu)\n", subNum+1, subLen);

		// keep track of how many attempts we've made to generate this
		// subrange;  after a while, we just assume the user has asked for
		// more subranges than can actually be provided given the underlying
		// data (e.g. the data contains too many invalid nucleotides)

		attempts = 0;

	try_again:
		if ((verbosity >= 20) && (attempts > 0))
			fprintf (stderr, "attempt %d\n", attempts+1);

		if (idTag != NULL) free (idTag);

		if (++attempts > maxAttempts)
			fatal ("unable to generate next subrange in reasonable time");

		// pick a starting location

		start = saveStart = urand32 (0, totalPlaces-1);

		// figure out where it is

		sub.start = 0;

		for (ix=0 ; ix<r->len ; ix++)
			{
			rng = &r->vec[ix];
			rngLen = (rng->end+1) - rng->start;
			if (rngLen < minLen) continue;
			places = rngLen+1 - minLen;
			if (start >= places)
				{ start -= places;  continue; }

			sub.chr     = rng->chr;
			sub.start   = rng->start + start;
			sub.end     = sub.start + subLen-1;
			sub.strand  = rng->strand; // (might modify strand info below)
			sub.extra   = NULL;
			sub.infoPtr = NULL;

			if (sub.end > rng->end)    // (intervals longer than the minimum
				{				       //  might cross end;  reject such case)
				if (verbosity >= 10)
					{
					range_fprintf (stderr, &sub, "%s %lu %lu %s %s");
					fprintf       (stderr, " too long\n");
					}
				goto try_again;
				}

			break;
			}

		if (sub.start == 0)
			fatalf ("internal error: unable to locate %ld\n", saveStart);

		idNum++;

		if (idTags)
			sub.extra = idTag = new_string_f ("id%dx", idNum);

		if (verbosity >= 10)
			range_fprintf (stderr, &sub, "%s %lu %lu %s %s");

		// make sure this subrange doesn't overlap with a previous one

		if (!allowOverlap)
			{
			ranges_sort (&subranges, NULL);
			if (range_in_ranges (&sub, &subranges))
				{
				if (verbosity >= 10) fprintf (stderr, " overlap\n");
				goto try_again;
				}
			}

		// fetch the nucleotides for this subrange (if we need 'em)

		if (seq != NULL) // close the nucleotide sequence
			{
			ckfree (seqFileName);  seqFileName = NULL;
			seq_close (seq);       seq         = NULL;
			}

		if ((!allowBadNucs) || (fastaF != NULL))
			{
			seqFileName = new_string_f
			                 ("%s%s%s[%ld,%ld]",
			                  genomeDir, chromosome_to_str(sub.chr),
			                  genomeExt, sub.start, sub.end);
			seq = seq_open_type (seqFileName, SEQ_TYPE_GUESS);
			if (!seq_read (seq))
				fatalf ("%s is empty", seqFileName);
			}

		// make sure this subrange doesn't contain bad nucleotides

		if (!allowBadNucs)
			{
			if (!valid_sequence (seq))
				{
				if (verbosity >= 10) fprintf (stderr, " invalid\n");
				goto try_again;
				}
			}

		// 'accept' this subrange

		if (verbosity >= 10) fprintf (stderr, " (#%d)\n", subNum+1);

		if (randomStrands)
			sub.strand = (urand32(0,1)==0)? '+' : '-';

		if (!allowOverlap)
			{
			if (!ranges_append (&subranges, &sub))
				fatal ("unable to keep track of subranges");
			}

		// show this subrange

		range_fprintf (stdout, &sub, "%s %lu %lu %s %s\n");

		if (fastaF != NULL)
			{
			range_fprintf  (fastaF, &sub, "> %s %lu %lu %s %s\n");
			print_as_fasta (fastaF, seq);
			}

		if (allowOverlap)
			{ free (idTag);  idTag = NULL; }
		else
			idTag = NULL; // (it has been copied to the subranges list)
		}

	// clean up

	ranges_fini (&subranges);

	if (seq != NULL) // close the nucleotide sequence
		{
		ckfree (seqFileName);  seqFileName = NULL;
		seq_close (seq);       seq         = NULL;
		}

	if (fastaF != NULL) fclose (fastaF);
	}

//----------
//
// valid_sequence--
//  Determine if a sequence is valid DNA.
//
//----------
//
// Arguments:
//  SEQ*	seq:	The sequence to check.
//
// Returns:
//  true if the sequence contains only valid DNA characters (ACGTacgt);  false
//  if it includes anything else.
//
//----------

static int valid_sequence
   (SEQ*	seq)
	{
	int		sLen = SEQ_LEN (seq);
	int		ix;

	for (ix=0 ; ix<sLen; ix++)
		{
		if (strchr ("ACGTacgt", seq->seq[ix]) == NULL)
			return false;
		}

	return true;
	}

//----------
//
// print_as_fasta--
//  Print a sequence in fasta format.
//
//----------
//
// Arguments:
//  FILE*	f:		The file to print to.
//  SEQ*	seq:	The sequence to print.
//
// Returns:
//  (nothing)
//
//----------

static void print_as_fasta
   (FILE*			f,
	SEQ*			seq)
	{
	int				sLen = SEQ_LEN (seq);
	int				lineLen, sIx, ix;
	unsigned long	lineIndex, lineEndIndex;

	//////////
	// print the sequence one line at a time
	//////////

	for (sIx=0 ; sIx<sLen; sIx+=charsPerLine)
		{
		lineIndex = seq->from + sIx;

		lineLen = charsPerLine;
		if (sIx + lineLen > sLen)
			lineLen = sLen - sIx;

		lineEndIndex = lineIndex+lineLen-1;

		// print a row of the sequence

		for (ix=0 ; ix<lineLen; ix++)
			fprintf (f, "%c", SEQ_AT (seq, sIx+ix));
		fprintf (f, "\n");
		}

	}

//----------
//
// create_lengths--
//  Create a set of lengths, all the same.
//
//----------
//
// Arguments:
//	int			howMany:	Number of lengths to create.
//	int			length:		Size of each length.
//	intvec_t*	lengths:	Place to create the lengths.  This must be a valid,
//							.. initialized intvec, and any other lengths it
//							.. contains will be discarded.
//
// Returns:
//  (nothing)
//
//----------

static void create_lengths
   (int			howMany,
	int			length,
	intvec_t*	lengths)
	{
	int			i;

	intvec_fini (lengths);

	for (i=0 ; i<howMany ; i++)
		{
		if (!intvec_append (lengths, length))
			fatalf ("unable to create length table for %d lengths", howMany);
		}

	intvec_fit (lengths);
	}

//----------
//
// ranges_to_lengths--
//  Copy the lengths of a list of ranges.
//
//----------
//
// Arguments:
//	ranges_t*	r:			The range list to take the lengths from.
//	intvec_t*	lengths:	Place to create the lengths.  This must be a valid,
//							.. initialized intvec, and any other lengths it
//							.. contains will be discarded.
//
// Returns:
//  (nothing)
//
//----------

static void ranges_to_lengths
   (ranges_t*	r,
	intvec_t*	lengths)
	{
	int			rIx;
	range_t		rng;

	intvec_fini (lengths);

	for (rIx=0 ; rIx<r->len ; rIx++)
		{
		rng = r->vec[rIx];

		if (!intvec_append (lengths, rng.end+1 - rng.start))
			fatalf ("unable to create length table for %d lengths", r->len);
		}

	intvec_fit (lengths);
	}

//----------
//
// read_lengths--
//  Read a list of ranges from a file.
//
//----------
//
// Arguments:
//	FILE*		f:			The file to read the lengths from;  it is assumed
//							.. to have been opened as text, and positioned at
//							.. the beginning.
//	intvec_t*	lengths:	Place to create the lengths.  This must be a valid,
//							.. initialized intvec, and any other lengths it
//							.. contains will be discarded.  If the file is not
//							.. of the proper format, the length of this list
//							.. will be zero.
//
// Returns:
//  (nothing)
//
//----------

static void read_lengths
   (FILE*		f,
	intvec_t*	lengths)
	{
	int			result;
	int			len;
	int			numLengths = 0;

	intvec_fini (lengths);

	while (true)
		{
		result = fscanf (f, "%d", &len);
		if (result == EOF) break;

		if (result != 1)
			{ intvec_fini (lengths);  return; }

		numLengths++;
		if (!intvec_append (lengths, len))
			fatalf ("unable to read length table for %d lengths", numLengths);
		}

	intvec_fit (lengths);
	}

//----------
//
// shuffle_lengths--
//  Randomly shuffle a list of ranges.
//
//----------
//
// Arguments:
//	intvec_t*	lengths:	The list of lengths to shuffle.
//
// Returns:
//  (nothing)
//
//----------

static void shuffle_lengths
   (intvec_t*	lengths)
	{
	int			ix, rIx;
	int			t;

	for (ix=0 ; ix<lengths->len-1 ; ix++)
		{
		rIx = urand32 (ix, lengths->len-1);
		if (rIx != ix)
			{
			t               = lengths->a[rIx];
			lengths->a[rIx] = lengths->a[ix];
			lengths->a[ix]  = t;
			}
		}

	}

//----------
//
// get_args--
//  Parse command line arguments.
//
//----------
//
// Arguments:
//	argc, argv: (as per main)
//
// Returns:
//  (nothing)
//
// Globals Set:
//	All options and feature descriptors.
//
//----------

static void usage (void)
	{
	fprintf (stderr, "grandom-- choose a set of random subintervals from a genome\n");
	fprintf (stderr, "usage: grandom [genome] [length] [count] [options]\n");
	fprintf (stderr, "  [genome]        File containing a list of intervals defining the allowed\n");
	fprintf (stderr, "                  portion of the genome;  intervals are inclusive on both ends\n");
	fprintf (stderr, "                  and chromosomes begin at position 1, e.g.\n");
	fprintf (stderr, "                    chr1 1 246127941\n");
	fprintf (stderr, "                    chr2 1 243615958\n");
	fprintf (stderr, "                     ...\n");
	fprintf (stderr, "                  (if no file specified, intervals are read from stdin)\n");
	fprintf (stderr, "  [length]        Number of base pairs in each random subinterval\n");
	fprintf (stderr, "                  (required, unless -l is used)\n");
	fprintf (stderr, "  [count]         Number of random subintervals to choose\n");
	fprintf (stderr, "                  (cannot be used with -l)\n");
	fprintf (stderr, "  [-x file]       File containing a list of excluded intervals\n");
	fprintf (stderr, "  [-l file]       File describing lengths of the random subintervals, either\n");
	fprintf (stderr, "                  whitespace separated numbers or a list of intervals\n");
	fprintf (stderr, "                  (replaces length and count arguments)\n");
	fprintf (stderr, "  [--overlapok]   Allow overlapping subintervals\n");
	fprintf (stderr, "  [--checknt]     Discard subintervals with invalid nucleotides\n");
	fprintf (stderr, "  [-d directory]  Directory containing nucleotide files\n");
	fprintf (stderr, "  [.ext]          File extension for nucleotide files (default is .nib)\n");
	fprintf (stderr, "                  (nucleotide files may be in nib of fasta format)\n");
	fprintf (stderr, "  [--strand]      Choose random strands (default is \"no strand\")\n");
	fprintf (stderr, "  [--idtags]      Assign a unique id tag to each subinterval\n");
	fprintf (stderr, "  [-f fasta]      Output sequences for the subintervals (to a fasta file)\n");
	fprintf (stderr, "  [-s seed]       Set the seed for the random number generator\n");
	fprintf (stderr, "  [-so file]      Output the seed to a file\n");
	fprintf (stderr, "  [-v verbosity]  Set info level (0 is minimum, 10 is everything)\n");
	exit(1);
	}

static void get_args
   (int			_argc,
	char**		_argv)
	{
	int			argc;
	char**		argv;
	int			argVal;
	time_t		t;
	int			seedWasSet;

	genomeName  = NULL;
	genomeDir   = NULL;
	genomeExt   = NULL;
	excludeName = NULL;
	lengthsName = NULL;

	// skip program name

	argv = _argv+1;  argc = _argc - 1;

	//////////
	// scan arguments
	//////////

	subrangeLength = 0;
	numSubranges   = 0;
	randomStrands  = false;
	allowOverlap   = false;
	allowBadNucs   = true;
	seedWasSet     = false;

	while (argc > 0)
		{
		// -strand

		if ((strcmp (argv[0], "-strand") == 0)
		 || (strcmp (argv[0], "--strand") == 0))
			{ randomStrands = true;  goto next_arg; }

		// -overlapok

		if ((strcmp (argv[0], "-overlapok") == 0)
		 || (strcmp (argv[0], "--overlapok") == 0))
			{ allowOverlap = true;  goto next_arg; }

		// -checknt

		if ((strcmp (argv[0], "-checknt") == 0)
		 || (strcmp (argv[0], "--checknt") == 0))
			{ allowBadNucs = false;  goto next_arg; }

		// -idtags

		if ((strcmp (argv[0], "-idtags") == 0)
		 || (strcmp (argv[0], "--idtags") == 0))
			{ idTags = true;  goto next_arg; }

		// -v verbosity

		if ((argc >= 2) && (strcmp (argv[0], "-v") == 0))
			{
			if (!string_to_int (argv[1], &verbosity))
				{
				fprintf (stderr, "%s %s is not valid\n", argv[0], argv[1]);
				usage ();
				}
			goto next_arg2;
			}

		// genome directory name

		if ((genomeDir == NULL)
		 && (argc >= 2) && (strcmp (argv[0], "-d") == 0))
			{
			if (argv[1][strlen(argv[1])-1] == '/')
				genomeDir = copy_string (argv[1]);
			else
				genomeDir = new_string_f ("%s/", argv[1]);
			goto next_arg2;
			}

		// genome files extension

		if ((genomeExt == NULL) && (argv[0][0] == '.'))
			{
			genomeExt = copy_string (argv[1]);
			goto next_arg2;
			}

		// excluded ranges name

		if ((excludeName == NULL)
		 && (argc >= 2) && (strcmp (argv[0], "-x") == 0))
			{ excludeName = copy_string (argv[1]);  goto next_arg2; }

		// lengths name

		if ((lengthsName == NULL)
		 && (argc >= 2) && (strcmp (argv[0], "-l") == 0))
			{ lengthsName = copy_string (argv[1]);  goto next_arg2; }

		// fasta file name

		if ((fastaName == NULL)
		 && (argc >= 2) && (strcmp (argv[0], "-f") == 0))
			{ fastaName = copy_string (argv[1]);  goto next_arg2; }

		// seed output file name

		if ((seedOutName == NULL)
		 && (argc >= 2) && (strcmp (argv[0], "-so") == 0))
			{ seedOutName = copy_string (argv[1]);  goto next_arg2; }

		// subrange length

		if (subrangeLength == 0)
			{
			if (string_to_int (argv[0], &subrangeLength))
				{
				if (subrangeLength < 0)
					{
					fprintf (stderr, "%d is not a valid subrange length\n",
									 subrangeLength);
					usage ();
					}
				goto next_arg;
				}
			// (keep going, this arg was not the subrange length)
			}

		// subrange count

		if (string_to_int (argv[0], &argVal))
			{
			if (argVal <= 0)
				{
				fprintf (stderr, "%d is not a valid number of subranges\n",
				                 numSubranges);
				usage ();
				}
			numSubranges = argVal;
			goto next_arg;
			}

		// random seed

		if ((argc >= 2) && (strcmp (argv[0], "-s") == 0))
			{
			int _randomSeed;
			if (!string_to_int (argv[1], &_randomSeed))
				{
				fprintf (stderr, "%s %s is not valid\n", argv[0], argv[1]);
				usage ();
				}
			randomSeed = _randomSeed;
			seedWasSet = true;
			goto next_arg2;
			}

		// genome ranges file name

		if (genomeName == NULL)
			{ genomeName = copy_string (argv[0]);  goto next_arg; }

		fprintf (stderr, "Can't understand \"%s\"\n", argv[0]);
		usage ();

	next_arg:
		argv++;  argc--;
		continue;

	next_arg2:
		argv+=2;  argc-=2;
		continue;
		}

	//////////
	// check for sensibility
	//////////

	if ((subrangeLength == 0) && (lengthsName == NULL))
		{
		fprintf (stderr, "You must at least specify a subrange length or lengths file\n");
		usage ();
		}

	if ((subrangeLength != 0) && (lengthsName != NULL))
		{
		fprintf (stderr, "You can't specify a subrange length together with a lengths file\n");
		usage ();
		}

	if ((numSubranges != 0) && (lengthsName != NULL))
		{
		fprintf (stderr, "You can't specify a subrange count together with a lengths file\n");
		usage ();
		}

	if ((genomeExt != NULL) && (genomeDir == NULL))
		{
		fprintf (stderr, "You can't specify a file extensions without a genome directory\n");
		usage ();
		}

	if ((!allowBadNucs) && (genomeDir == NULL))
		{
		fprintf (stderr, "With -checknt you must at least specify a genome directory\n");
		usage ();
		}

	if ((fastaName != NULL) && (genomeDir == NULL))
		{
		fprintf (stderr, "With -f you must at least specify a genome directory\n");
		usage ();
		}

	if (numSubranges == 0)
		numSubranges = 1;

	if ((genomeDir != NULL) && (genomeExt == NULL))
		genomeExt = copy_string (".nib");

	//////////
	// initialize the random number generator
	//////////

	if (!seedWasSet)
		{ time (&t);  randomSeed = t; }
	srand32 (randomSeed);
	}

//----------
//
// show_options--
//  Show the options that were set by the command line arguments.
//
//----------
//
// Arguments:
//  FILE*   f:  The stream to write to.
//
// Returns:
//  (nothing)
//
//----------

static void show_options
   (FILE*  f)
	{
	fprintf (f, "verbosity      = %d\n",  verbosity);
	fprintf (f, "subrangeLength = %d\n",  subrangeLength);
	fprintf (f, "numSubranges   = %d\n",  numSubranges);
	fprintf (f, "randomStrands  = %s\n",  (randomStrands)? "true" : "false");
	fprintf (f, "randomSeed     = %ld\n", randomSeed);
	fprintf (f, "allowOverlap   = %s\n",  (allowOverlap)?  "true" : "false");
	fprintf (f, "allowBadNucs   = %s\n",  (allowBadNucs)?  "true" : "false");
	fprintf (f, "idTags         = %s\n",  (idTags)?        "true" : "false");

	if (genomeName == NULL)
		fprintf (f, "genomeName     = (null)\n");
	else
		fprintf (f, "genomeName     = \"%s\"\n", genomeName);

	if (genomeDir == NULL)
		fprintf (f, "genomeDir      = (null)\n");
	else
		fprintf (f, "genomeDir      = \"%s\"\n", genomeDir);

	if (genomeExt == NULL)
		fprintf (f, "genomeExt      = (null)\n");
	else
		fprintf (f, "genomeExt      = \"%s\"\n", genomeExt);

	if (excludeName == NULL)
		fprintf (f, "excludeName    = (null)\n");
	else
		fprintf (f, "excludeName    = \"%s\"\n", excludeName);

	if (lengthsName == NULL)
		fprintf (f, "lengthsName    = (null)\n");
	else
		fprintf (f, "lengthsName    = \"%s\"\n", lengthsName);

	if (fastaName == NULL)
		fprintf (f, "fastaName      = (null)\n");
	else
		fprintf (f, "fastaName      = \"%s\"\n", fastaName);

	if (seedOutName == NULL)
		fprintf (f, "seedOutName    = (null)\n");
	else
		fprintf (f, "seedOutName    = \"%s\"\n", seedOutName);
	}

